/********************************
 Ben Ernest-Jones
 Markus Adhiwiyogo

 18-796

 DCT matrix operations

 ********************************/

#include <math.h>
#include "include/define.h"
#include "include/matrix.h"

void computeC(pel_t *C[],int size)
{
  pel_t k;
  int m,n;

  for(m=0;m<size;m++){
    for(n=0;n<size;n++){
      if (0==n)
	k=sqrt(1/(pel_t)size);
      else
	k=sqrt(2/(pel_t)size);
      C[m][n] = k * cos( ((2*(pel_t)m+1)*(pel_t)n*M_PI) / (2*(pel_t)size) );
    }//for n
  }//for m
}//computeC

void DCT(pel_t *Y[],pel_t *C[],pel_t *X[],int size)
{/* Discrete Cosine Transform:  Y = transpose(C) * X * C
  */

  int m,n,i;
  pel_t **tmp = new (pel_t *)[size];
  
  /*  tmp = transpose(C) * X  */
  for(m=0;m<size;m++){
    tmp[m] = new pel_t[size];
    for(n=0;n<size;n++){
      tmp[m][n]=0.0;
      for(i=0;i<size;i++)
	tmp[m][n]=tmp[m][n]+transpose(C,m,i)*X[i][n];
    }
  }

  /* Y = tmp *  C */
  for(m=0;m<=size-1;m++){
    for(n=0;n<=size-1;n++){
      Y[m][n]=0.0;
      for(i=0;i<=size-1;i++)
	Y[m][n]=Y[m][n]+tmp[m][i]*C[i][n];
    }
    delete tmp[m];
  }
  delete tmp;
}


void invDCT(pel_t *X[],pel_t *C[],pel_t *Y[],int size)
{/* Inverse Discrete Cosine Transform:  X = C * Y * transpose(C)
  */

  int m,n,i;
  pel_t **tmp = new (pel_t *)[size];

  /*  tmp = C * Y  */
  for(m=0;m<size;m++){
    tmp[m] = new pel_t[size];
    for(n=0;n<size;n++){
      tmp[m][n]=0.0;
      for(i=0;i<=size-1;i++)
	tmp[m][n]=tmp[m][n]+C[m][i]*Y[i][n];
    }
  }

  /* X = tmp *  transpose(C) */
  for(m=0;m<=size-1;m++){
    for(n=0;n<=size-1;n++){
      X[m][n]=0.0;
      for(i=0;i<=size-1;i++)
	X[m][n]=X[m][n]+tmp[m][i]*transpose(C,i,n);
      X[m][n]=clip(X[m][n],-256,255);
    }
    delete tmp[m];
  }
  delete tmp;
}
